# coding: utf-8

"""
    Coinbase Developer Platform APIs

    The Coinbase Developer Platform APIs - leading the world's transition onchain.

    The version of the OpenAPI document: 2.0.0
    Contact: cdp@coinbase.com
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501

import warnings
from pydantic import validate_call, Field, StrictFloat, StrictStr, StrictInt
from typing import Any, Dict, List, Optional, Tuple, Union
from typing_extensions import Annotated

from pydantic import Field, StrictStr, field_validator
from typing import List, Optional
from typing_extensions import Annotated
from cdp.openapi_client.models.create_payment_transfer_quote201_response import CreatePaymentTransferQuote201Response
from cdp.openapi_client.models.create_payment_transfer_quote_request import CreatePaymentTransferQuoteRequest
from cdp.openapi_client.models.crypto_rail import CryptoRail
from cdp.openapi_client.models.payment_method import PaymentMethod
from cdp.openapi_client.models.transfer import Transfer

from cdp.openapi_client.api_client import ApiClient, RequestSerialized
from cdp.openapi_client.api_response import ApiResponse
from cdp.openapi_client.rest import RESTResponseType


class PaymentsAlphaApi:
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None) -> None:
        if api_client is None:
            api_client = ApiClient.get_default()
        self.api_client = api_client


    @validate_call
    async def create_payment_transfer_quote(
        self,
        create_payment_transfer_quote_request: CreatePaymentTransferQuoteRequest,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> CreatePaymentTransferQuote201Response:
        """Create a transfer quote

        Creates a new transfer quote, which can then be executed using the Execute a transfer quote endpoint. If you want to automatically execute the transfer without needing to confirm, specify execute as true.

        :param create_payment_transfer_quote_request: (required)
        :type create_payment_transfer_quote_request: CreatePaymentTransferQuoteRequest
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._create_payment_transfer_quote_serialize(
            create_payment_transfer_quote_request=create_payment_transfer_quote_request,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '201': "CreatePaymentTransferQuote201Response",
            '400': "Error",
            '401': "Error",
            '429': "Error",
            '500': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    async def create_payment_transfer_quote_with_http_info(
        self,
        create_payment_transfer_quote_request: CreatePaymentTransferQuoteRequest,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[CreatePaymentTransferQuote201Response]:
        """Create a transfer quote

        Creates a new transfer quote, which can then be executed using the Execute a transfer quote endpoint. If you want to automatically execute the transfer without needing to confirm, specify execute as true.

        :param create_payment_transfer_quote_request: (required)
        :type create_payment_transfer_quote_request: CreatePaymentTransferQuoteRequest
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._create_payment_transfer_quote_serialize(
            create_payment_transfer_quote_request=create_payment_transfer_quote_request,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '201': "CreatePaymentTransferQuote201Response",
            '400': "Error",
            '401': "Error",
            '429': "Error",
            '500': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    async def create_payment_transfer_quote_without_preload_content(
        self,
        create_payment_transfer_quote_request: CreatePaymentTransferQuoteRequest,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Create a transfer quote

        Creates a new transfer quote, which can then be executed using the Execute a transfer quote endpoint. If you want to automatically execute the transfer without needing to confirm, specify execute as true.

        :param create_payment_transfer_quote_request: (required)
        :type create_payment_transfer_quote_request: CreatePaymentTransferQuoteRequest
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._create_payment_transfer_quote_serialize(
            create_payment_transfer_quote_request=create_payment_transfer_quote_request,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '201': "CreatePaymentTransferQuote201Response",
            '400': "Error",
            '401': "Error",
            '429': "Error",
            '500': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _create_payment_transfer_quote_serialize(
        self,
        create_payment_transfer_quote_request,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[
            str, Union[str, bytes, List[str], List[bytes], List[Tuple[str, bytes]]]
        ] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter
        if create_payment_transfer_quote_request is not None:
            _body_params = create_payment_transfer_quote_request


        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                [
                    'application/json'
                ]
            )

        # set the HTTP header `Content-Type`
        if _content_type:
            _header_params['Content-Type'] = _content_type
        else:
            _default_content_type = (
                self.api_client.select_header_content_type(
                    [
                        'application/json'
                    ]
                )
            )
            if _default_content_type is not None:
                _header_params['Content-Type'] = _default_content_type

        # authentication setting
        _auth_settings: List[str] = [
            'apiKeyAuth'
        ]

        return self.api_client.param_serialize(
            method='POST',
            resource_path='/v2/payments/transfers',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )




    @validate_call
    async def execute_payment_transfer_quote(
        self,
        transfer_id: Annotated[str, Field(strict=True, description="The ID of the transfer.")],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> Transfer:
        """Execute a transfer quote

        Executes a transfer quote which was created using the Create a transfer quote endpoint.

        :param transfer_id: The ID of the transfer. (required)
        :type transfer_id: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._execute_payment_transfer_quote_serialize(
            transfer_id=transfer_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "Transfer",
            '400': "Error",
            '401': "Error",
            '404': "Error",
            '429': "Error",
            '500': "Error",
            '502': "Error",
            '503': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    async def execute_payment_transfer_quote_with_http_info(
        self,
        transfer_id: Annotated[str, Field(strict=True, description="The ID of the transfer.")],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[Transfer]:
        """Execute a transfer quote

        Executes a transfer quote which was created using the Create a transfer quote endpoint.

        :param transfer_id: The ID of the transfer. (required)
        :type transfer_id: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._execute_payment_transfer_quote_serialize(
            transfer_id=transfer_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "Transfer",
            '400': "Error",
            '401': "Error",
            '404': "Error",
            '429': "Error",
            '500': "Error",
            '502': "Error",
            '503': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    async def execute_payment_transfer_quote_without_preload_content(
        self,
        transfer_id: Annotated[str, Field(strict=True, description="The ID of the transfer.")],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Execute a transfer quote

        Executes a transfer quote which was created using the Create a transfer quote endpoint.

        :param transfer_id: The ID of the transfer. (required)
        :type transfer_id: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._execute_payment_transfer_quote_serialize(
            transfer_id=transfer_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "Transfer",
            '400': "Error",
            '401': "Error",
            '404': "Error",
            '429': "Error",
            '500': "Error",
            '502': "Error",
            '503': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _execute_payment_transfer_quote_serialize(
        self,
        transfer_id,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[
            str, Union[str, bytes, List[str], List[bytes], List[Tuple[str, bytes]]]
        ] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        if transfer_id is not None:
            _path_params['transferId'] = transfer_id
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter


        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                [
                    'application/json'
                ]
            )


        # authentication setting
        _auth_settings: List[str] = [
            'apiKeyAuth'
        ]

        return self.api_client.param_serialize(
            method='POST',
            resource_path='/v2/payments/transfers/{transferId}/execute',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )




    @validate_call
    async def get_crypto_rails(
        self,
        networks: Annotated[Optional[StrictStr], Field(description="Comma separated list of networks to filter the rails by.")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> List[CryptoRail]:
        """Get the crypto rails

        Gets the crypto rails that can be used to send funds or receive funds.

        :param networks: Comma separated list of networks to filter the rails by.
        :type networks: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._get_crypto_rails_serialize(
            networks=networks,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "List[CryptoRail]",
            '401': "Error",
            '500': "Error",
            '502': "Error",
            '503': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    async def get_crypto_rails_with_http_info(
        self,
        networks: Annotated[Optional[StrictStr], Field(description="Comma separated list of networks to filter the rails by.")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[List[CryptoRail]]:
        """Get the crypto rails

        Gets the crypto rails that can be used to send funds or receive funds.

        :param networks: Comma separated list of networks to filter the rails by.
        :type networks: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._get_crypto_rails_serialize(
            networks=networks,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "List[CryptoRail]",
            '401': "Error",
            '500': "Error",
            '502': "Error",
            '503': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    async def get_crypto_rails_without_preload_content(
        self,
        networks: Annotated[Optional[StrictStr], Field(description="Comma separated list of networks to filter the rails by.")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Get the crypto rails

        Gets the crypto rails that can be used to send funds or receive funds.

        :param networks: Comma separated list of networks to filter the rails by.
        :type networks: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._get_crypto_rails_serialize(
            networks=networks,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "List[CryptoRail]",
            '401': "Error",
            '500': "Error",
            '502': "Error",
            '503': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _get_crypto_rails_serialize(
        self,
        networks,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[
            str, Union[str, bytes, List[str], List[bytes], List[Tuple[str, bytes]]]
        ] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        if networks is not None:
            
            _query_params.append(('networks', networks))
            
        # process the header parameters
        # process the form parameters
        # process the body parameter


        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                [
                    'application/json'
                ]
            )


        # authentication setting
        _auth_settings: List[str] = [
            'apiKeyAuth'
        ]

        return self.api_client.param_serialize(
            method='GET',
            resource_path='/v2/payments/rails/crypto',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )




    @validate_call
    async def get_payment_methods(
        self,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> List[PaymentMethod]:
        """Get the fiat payment methods

        Gets the fiat payment methods that can be used to send funds or receive funds. This is the list of payment methods configured for your account.

        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._get_payment_methods_serialize(
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "List[PaymentMethod]",
            '401': "Error",
            '500': "Error",
            '502': "Error",
            '503': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    async def get_payment_methods_with_http_info(
        self,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[List[PaymentMethod]]:
        """Get the fiat payment methods

        Gets the fiat payment methods that can be used to send funds or receive funds. This is the list of payment methods configured for your account.

        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._get_payment_methods_serialize(
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "List[PaymentMethod]",
            '401': "Error",
            '500': "Error",
            '502': "Error",
            '503': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    async def get_payment_methods_without_preload_content(
        self,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Get the fiat payment methods

        Gets the fiat payment methods that can be used to send funds or receive funds. This is the list of payment methods configured for your account.

        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._get_payment_methods_serialize(
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "List[PaymentMethod]",
            '401': "Error",
            '500': "Error",
            '502': "Error",
            '503': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _get_payment_methods_serialize(
        self,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[
            str, Union[str, bytes, List[str], List[bytes], List[Tuple[str, bytes]]]
        ] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter


        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                [
                    'application/json'
                ]
            )


        # authentication setting
        _auth_settings: List[str] = [
            'apiKeyAuth'
        ]

        return self.api_client.param_serialize(
            method='GET',
            resource_path='/v2/payments/rails/payment-methods',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )




    @validate_call
    async def get_payment_transfer(
        self,
        transfer_id: Annotated[str, Field(strict=True, description="The ID of the transfer.")],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> Transfer:
        """Get a transfer by ID

        Gets a transfer by ID.

        :param transfer_id: The ID of the transfer. (required)
        :type transfer_id: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._get_payment_transfer_serialize(
            transfer_id=transfer_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "Transfer",
            '400': "Error",
            '401': "Error",
            '404': "Error",
            '500': "Error",
            '502': "Error",
            '503': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    async def get_payment_transfer_with_http_info(
        self,
        transfer_id: Annotated[str, Field(strict=True, description="The ID of the transfer.")],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[Transfer]:
        """Get a transfer by ID

        Gets a transfer by ID.

        :param transfer_id: The ID of the transfer. (required)
        :type transfer_id: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._get_payment_transfer_serialize(
            transfer_id=transfer_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "Transfer",
            '400': "Error",
            '401': "Error",
            '404': "Error",
            '500': "Error",
            '502': "Error",
            '503': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    async def get_payment_transfer_without_preload_content(
        self,
        transfer_id: Annotated[str, Field(strict=True, description="The ID of the transfer.")],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Get a transfer by ID

        Gets a transfer by ID.

        :param transfer_id: The ID of the transfer. (required)
        :type transfer_id: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._get_payment_transfer_serialize(
            transfer_id=transfer_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "Transfer",
            '400': "Error",
            '401': "Error",
            '404': "Error",
            '500': "Error",
            '502': "Error",
            '503': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _get_payment_transfer_serialize(
        self,
        transfer_id,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[
            str, Union[str, bytes, List[str], List[bytes], List[Tuple[str, bytes]]]
        ] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        if transfer_id is not None:
            _path_params['transferId'] = transfer_id
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter


        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                [
                    'application/json'
                ]
            )


        # authentication setting
        _auth_settings: List[str] = [
            'apiKeyAuth'
        ]

        return self.api_client.param_serialize(
            method='GET',
            resource_path='/v2/payments/transfers/{transferId}',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )





# coding: utf-8

"""
    Coinbase Developer Platform APIs

    The Coinbase Developer Platform APIs - leading the world's transition onchain.

    The version of the OpenAPI document: 2.0.0
    Contact: cdp@coinbase.com
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


import unittest

from cdp.openapi_client.models.token_fee import TokenFee

class TestTokenFee(unittest.TestCase):
    """TokenFee unit test stubs"""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def make_instance(self, include_optional) -> TokenFee:
        """Test TokenFee
            include_optional is a boolean, when False only required
            params are included, when True both required and
            optional params are included """
        # uncomment below to create an instance of `TokenFee`
        """
        model = TokenFee()
        if include_optional:
            return TokenFee(
                amount = '1000000000000000000',
                token = '0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE'
            )
        else:
            return TokenFee(
                amount = '1000000000000000000',
                token = '0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE',
        )
        """

    def testTokenFee(self):
        """Test TokenFee"""
        # inst_req_only = self.make_instance(include_optional=False)
        # inst_req_and_optional = self.make_instance(include_optional=True)

if __name__ == '__main__':
    unittest.main()



import time

from cdp.api_clients import ApiClients


async def wait_for_fund_operation_receipt(
    api_clients: ApiClients,
    transfer_id: str,
    timeout_seconds: float = 900,
    interval_seconds: float = 1,
):
    """Wait for a fund operation to be processed.

    Args:
        api_clients: The API clients object.
        transfer_id (str): The id of the transfer to wait for.
        timeout_seconds (float, optional): Maximum time to wait in seconds. Defaults to 20.
        interval_seconds (float, optional): Time between checks in seconds. Defaults to 0.2.

    Returns:
        Transfer: The final transfer object.

    Raises:
        TimeoutError: If the operation doesn't complete within the specified timeout.

    """
    start_time = time.time()

    # Get initial state - this matches what we see in the diff
    transfer = await api_clients.payments.get_payment_transfer(
        transfer_id,
    )

    # Use a regular while loop that explicitly checks the status
    while transfer.status not in ["completed", "failed"]:
        # Check timeout before making next API call
        if time.time() - start_time > timeout_seconds:
            raise TimeoutError("Transfer timed out")

        # Wait before checking again
        time.sleep(interval_seconds)

        # Make API call to check status
        transfer = await api_clients.payments.get_payment_transfer(
            transfer_id,
        )

    return transfer



import inspect
import re

import pytest

from cdp.evm_client import EvmClient
from cdp.openapi_client.api.evm_accounts_api import EVMAccountsApi
from cdp.openapi_client.api.evm_smart_accounts_api import EVMSmartAccountsApi
from cdp.openapi_client.api.evm_token_balances_api import EVMTokenBalancesApi
from cdp.openapi_client.api.faucets_api import FaucetsApi
from cdp.openapi_client.api.solana_accounts_api import SolanaAccountsApi
from cdp.solana_client import SolanaClient


def get_public_methods(cls):
    """Extract public methods from a class that aren't special methods."""
    methods = []
    for name, _ in inspect.getmembers(cls, inspect.isfunction):
        # Skip private, special methods, and methods with _serialize, _with_http_info, or _without_preload_content suffixes
        if not name.startswith("_") and not re.search(
            r"_serialize$|_with_http_info$|_without_preload_content$", name
        ):
            methods.append(name)
    return methods


def is_method_wrapped(api_method, evm_methods, solana_methods):
    """Check if an API method is wrapped in either client."""
    if api_method in evm_methods or api_method in solana_methods:
        return True

    # For methods with 'evm' or 'solana', check for wrapped names without the evm or solana prefix
    stripped_method = api_method
    if "evm" in api_method:
        stripped_method = api_method.replace("evm_", "").replace("_by_name", "")
        if stripped_method in evm_methods:
            return True
    elif "solana" in api_method:
        stripped_method = api_method.replace("solana_", "").replace("_by_name", "")
        if stripped_method in solana_methods:
            return True
    elif "sol" in api_method:
        stripped_method = api_method.replace("sol_", "").replace("_by_name", "")
        if stripped_method in solana_methods:
            return True
    return False


@pytest.mark.asyncio
async def test_cdp_client_wraps_all_api_methods():
    """Test that client classes properly wrap all API methods."""
    evm_methods = get_public_methods(EvmClient)
    solana_methods = get_public_methods(SolanaClient)

    api_classes = [
        EVMAccountsApi,
        EVMSmartAccountsApi,
        EVMTokenBalancesApi,
        FaucetsApi,
        SolanaAccountsApi,
    ]

    for api_class in api_classes:
        api_methods = get_public_methods(api_class)

        for api_method in api_methods:
            wrapped = is_method_wrapped(api_method, evm_methods, solana_methods)

            print(f"Checking {api_method}: {'✓' if wrapped else '✗'}")

            assert wrapped, (
                f"API method '{api_method}' from {api_class.__name__} is not wrapped in either client. "
                f"Add a wrapper function to the appropriate client."
            )



from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from eth_account.messages import _hash_eip191_message, encode_defunct
from eth_account.typed_transactions import DynamicFeeTransaction
from eth_typing import Hash32
from hexbytes import HexBytes
from web3 import Web3

from cdp.actions.evm.fund.quote import Quote
from cdp.api_clients import ApiClients
from cdp.evm_server_account import EvmServerAccount
from cdp.evm_token_balances import (
    EvmToken,
    EvmTokenAmount,
    EvmTokenBalance,
    ListTokenBalancesResult,
)
from cdp.evm_transaction_types import TransactionRequestEIP1559
from cdp.openapi_client.models.create_payment_transfer_quote201_response import (
    CreatePaymentTransferQuote201Response,
)
from cdp.openapi_client.models.eip712_domain import EIP712Domain
from cdp.openapi_client.models.eip712_message import EIP712Message
from cdp.openapi_client.models.request_evm_faucet_request import RequestEvmFaucetRequest
from cdp.openapi_client.models.send_evm_transaction200_response import SendEvmTransaction200Response
from cdp.openapi_client.models.send_evm_transaction_request import SendEvmTransactionRequest
from cdp.openapi_client.models.sign_evm_hash_request import SignEvmHashRequest
from cdp.openapi_client.models.sign_evm_message_request import SignEvmMessageRequest
from cdp.openapi_client.models.sign_evm_transaction_request import (
    SignEvmTransactionRequest,
)


@patch("cdp.evm_server_account.EVMAccountsApi")
def test_initialization(mock_api, server_account_model_factory):
    """Test that the EvmServerAccount initializes correctly."""
    address = "0x1234567890123456789012345678901234567890"
    name = "test-account"

    server_account_model = server_account_model_factory(address, name)
    server_account = EvmServerAccount(server_account_model, mock_api, mock_api)

    assert server_account.address == address
    assert server_account.name == name


@patch("cdp.evm_server_account.EVMAccountsApi")
def test_str_representation(mock_api, server_account_model_factory):
    """Test the string representation of EvmServerAccount."""
    address = "0x1234567890123456789012345678901234567890"

    server_account_model = server_account_model_factory(address)
    server_account = EvmServerAccount(server_account_model, mock_api, mock_api)

    expected = f"Ethereum Account Address: {address}"
    assert str(server_account) == expected


@patch("cdp.evm_server_account.EVMAccountsApi")
def test_repr_representation(mock_api, server_account_model_factory):
    """Test the repr representation of EvmServerAccount."""
    address = "0x1234567890123456789012345678901234567890"
    server_account_model = server_account_model_factory(address)
    server_account = EvmServerAccount(server_account_model, mock_api, mock_api)

    expected = f"Ethereum Account Address: {address}"
    assert repr(server_account) == expected


@pytest.mark.asyncio
@patch("cdp.evm_server_account.EVMAccountsApi")
async def test_sign_message_with_bytes(mock_api, server_account_model_factory):
    """Test sign_message method with bytes message."""
    address = "0x1234567890123456789012345678901234567890"
    name = "test-account"
    server_account_model = server_account_model_factory(address, name)

    # Create a real mock instance, not just the class
    mock_api_instance = mock_api.return_value
    server_account = EvmServerAccount(server_account_model, mock_api_instance, mock_api_instance)

    message = b"Test message"
    signable_message = encode_defunct(message)

    signature_response = AsyncMock()
    # Create a real bytes-like object for the signature that's 65 bytes long
    # (32 bytes for r, 32 bytes for s, 1 byte for v). 1b = 27 in hex
    mock_signature = bytes.fromhex("1234" * 32 + "5678" * 32 + "1b")

    signature_response.signature = mock_signature
    mock_api_instance.sign_evm_message = AsyncMock(return_value=signature_response)

    result = await server_account.sign_message(signable_message)

    message_hex = HexBytes(message).hex()
    sign_request = SignEvmMessageRequest(message=message_hex)
    mock_api_instance.sign_evm_message.assert_called_once_with(
        address=address,
        sign_evm_message_request=sign_request,
        x_idempotency_key=None,
    )

    assert result.r == int.from_bytes(mock_signature[0:32], byteorder="big")
    assert result.s == int.from_bytes(mock_signature[32:64], byteorder="big")
    assert result.v == mock_signature[64]
    assert result.signature == HexBytes(mock_signature)
    assert result.message_hash == _hash_eip191_message(signable_message)


@pytest.mark.asyncio
@patch("cdp.evm_server_account.EVMAccountsApi")
async def test_sign_typed_data(mock_api, server_account_model_factory):
    """Test sign_typed_data method."""
    address = "0x1234567890123456789012345678901234567890"

    server_account_model = server_account_model_factory(address)
    server_account = EvmServerAccount(server_account_model, mock_api, mock_api)

    domain = EIP712Domain(
        name="EIP712Domain",
        version="1",
        chain_id=1,
        verifying_contract="0x1234567890123456789012345678901234567890",
    )
    types = {
        "EIP712Domain": [
            {"name": "name", "type": "string"},
            {"name": "chainId", "type": "uint256"},
            {"name": "verifyingContract", "type": "address"},
        ],
    }
    primary_type = "EIP712Domain"
    message = {
        "name": "EIP712Domain",
        "chainId": 1,
        "verifyingContract": "0x1234567890123456789012345678901234567890",
    }
    test_idempotency_key = "test-idempotency-key"

    signature_response = AsyncMock()
    signature_response.signature = "0xsignature"
    mock_api.sign_evm_typed_data = AsyncMock(return_value=signature_response)

    result = await server_account.sign_typed_data(
        domain=domain,
        types=types,
        primary_type=primary_type,
        message=message,
        idempotency_key=test_idempotency_key,
    )

    assert result == signature_response.signature

    mock_api.sign_evm_typed_data.assert_called_once_with(
        address=address,
        eip712_message=EIP712Message(
            domain=domain,
            types=types,
            primary_type=primary_type,
            message=message,
        ),
        x_idempotency_key=test_idempotency_key,
    )


@pytest.mark.asyncio
@patch("cdp.evm_server_account.EVMAccountsApi")
async def test_unsafe_sign_hash(mock_api, server_account_model_factory):
    """Test unsafe_sign_hash method."""
    address = "0x1234567890123456789012345678901234567890"
    name = "test-account"
    server_account_model = server_account_model_factory(address, name)

    mock_api_instance = mock_api.return_value
    server_account = EvmServerAccount(server_account_model, mock_api_instance, mock_api_instance)

    message_hash = Hash32(bytes.fromhex("abcd" * 16))

    signature_response = MagicMock()
    # Create a real bytes-like object for the signature (65 bytes: 32 for r, 32 for s, 1 for v)
    mock_signature = bytes.fromhex("1234" * 32 + "5678" * 32 + "1b")
    signature_response.signature = mock_signature
    mock_api_instance.sign_evm_hash = AsyncMock(return_value=signature_response)

    result = await server_account.unsafe_sign_hash(message_hash)

    hash_hex = HexBytes(message_hash).hex()
    sign_request = SignEvmHashRequest(hash=hash_hex)
    mock_api_instance.sign_evm_hash.assert_called_once_with(
        address=address, sign_evm_hash_request=sign_request, x_idempotency_key=None
    )

    assert result.r == int.from_bytes(mock_signature[0:32], byteorder="big")
    assert result.s == int.from_bytes(mock_signature[32:64], byteorder="big")
    assert result.v == mock_signature[64]
    assert result.signature == HexBytes(mock_signature)
    assert result.message_hash == message_hash


@pytest.mark.asyncio
@patch("cdp.evm_server_account.Web3")
@patch("cdp.evm_server_account.TypedTransaction")
@patch("cdp.evm_server_account.EVMAccountsApi")
async def test_sign_transaction(mock_api, mock_typed_tx, mock_web3, server_account_model_factory):
    """Test sign_transaction method."""
    address = "0x1234567890123456789012345678901234567890"
    name = "test-account"
    server_account_model = server_account_model_factory(address, name)

    mock_api_instance = mock_api.return_value
    server_account = EvmServerAccount(server_account_model, mock_api_instance, mock_api_instance)

    transaction_dict = {
        "maxFeePerGas": 2000000000,
        "maxPriorityFeePerGas": 1000000000,
        "nonce": 0,
        "gas": 21000,
        "to": "0x2B5AD5c4795c026514f8317c7a215E218DcCD6cF",
        "value": 1000000000000000000,
        "data": "0x",
        "chainId": 1,
        "type": 2,
    }

    mock_tx_instance = MagicMock()
    mock_typed_tx.from_dict.return_value = mock_tx_instance

    mock_tx_instance.transaction_type = 2
    mock_tx_instance.transaction = MagicMock()
    mock_tx_instance.transaction.dictionary = {
        "nonce": 0,
        "maxFeePerGas": 2000000000,
        "maxPriorityFeePerGas": 1000000000,
        "gas": 21000,
        "to": "0x2B5AD5c4795c026514f8317c7a215E218DcCD6cF",
        "value": 1000000000000000000,
        "data": "0x",
        "chainId": 1,
        "v": 0,
        "r": 0,
        "s": 0,
    }
    mock_payload = bytes.fromhex("f864")
    mock_tx_instance.transaction.payload.return_value = mock_payload

    serialized_tx_with_type = bytes([2]) + mock_payload
    tx_hex_with_0x = "0x" + serialized_tx_with_type.hex()

    mock_signature = bytes.fromhex("1234" * 32 + "5678" * 32 + "1b")
    signature_response = MagicMock()
    signature_response.signed_transaction = mock_signature
    mock_api_instance.sign_evm_transaction = AsyncMock(return_value=signature_response)

    mock_tx_hash = bytes.fromhex("abcd" * 8)
    mock_web3.keccak.return_value = mock_tx_hash

    result = await server_account.sign_transaction(transaction_dict)

    mock_typed_tx.from_dict.assert_called_once_with(transaction_dict)
    assert mock_tx_instance.transaction.dictionary["v"] == 0
    assert mock_tx_instance.transaction.dictionary["r"] == 0
    assert mock_tx_instance.transaction.dictionary["s"] == 0
    mock_tx_instance.transaction.payload.assert_called_once()

    expected_request = SignEvmTransactionRequest(transaction=tx_hex_with_0x)
    mock_api_instance.sign_evm_transaction.assert_called_once_with(
        address=address,
        sign_evm_transaction_request=expected_request,
        x_idempotency_key=None,
    )

    mock_web3.keccak.assert_called_once_with(HexBytes(mock_signature))

    assert result.raw_transaction == HexBytes(mock_signature)
    assert result.hash == mock_tx_hash
    assert result.r == int.from_bytes(mock_signature[0:32], byteorder="big")
    assert result.s == int.from_bytes(mock_signature[32:64], byteorder="big")
    assert result.v == mock_signature[64]


@pytest.mark.asyncio
async def test_request_faucet(server_account_model_factory):
    """Test request_faucet method."""
    address = "0x1234567890123456789012345678901234567890"
    name = "test-account"
    server_account_model = server_account_model_factory(address, name)

    mock_faucets_api = AsyncMock()
    mock_api_instance = AsyncMock()
    mock_api_instance.faucets = mock_faucets_api

    mock_response = AsyncMock()
    mock_response.transaction_hash = "0x123"
    mock_faucets_api.request_evm_faucet = AsyncMock(return_value=mock_response)
    server_account = EvmServerAccount(server_account_model, mock_api_instance, mock_api_instance)

    result = await server_account.request_faucet(network="base-sepolia", token="eth")

    mock_faucets_api.request_evm_faucet.assert_called_once_with(
        request_evm_faucet_request=RequestEvmFaucetRequest(
            network="base-sepolia",
            token="eth",
            address=address,
        ),
    )
    assert result == "0x123"


@pytest.mark.asyncio
@patch("cdp.evm_server_account.EVMAccountsApi")
async def test_send_transaction_serialized(mock_api, server_account_model_factory):
    """Test send_transaction method with serialized transaction."""
    address = "0x1234567890123456789012345678901234567890"
    name = "test-account"
    server_account_model = server_account_model_factory(address, name)
    mock_api_instance = mock_api.return_value
    mock_api_instance.send_evm_transaction = AsyncMock(
        return_value=SendEvmTransaction200Response(transaction_hash="0x123")
    )
    server_account = EvmServerAccount(server_account_model, mock_api_instance, mock_api_instance)

    test_network = "base-sepolia"
    test_transaction = "0xabcdef"

    result = await server_account.send_transaction(
        network=test_network,
        transaction=test_transaction,
    )

    mock_api_instance.send_evm_transaction.assert_called_once_with(
        address=server_account.address,
        send_evm_transaction_request=SendEvmTransactionRequest(
            transaction=test_transaction, network=test_network
        ),
        x_idempotency_key=None,
    )
    assert result == "0x123"


@pytest.mark.asyncio
@patch("cdp.evm_server_account.EVMAccountsApi")
async def test_send_transaction_eip1559(mock_api, server_account_model_factory):
    """Test send_transaction method with EIP-1559 transaction."""
    address = "0x1234567890123456789012345678901234567890"
    name = "test-account"
    server_account_model = server_account_model_factory(address, name)
    mock_api_instance = mock_api.return_value
    mock_api_instance.send_evm_transaction = AsyncMock(
        return_value=SendEvmTransaction200Response(transaction_hash="0x456")
    )
    server_account = EvmServerAccount(server_account_model, mock_api_instance, mock_api_instance)

    w3 = Web3()

    test_network = "base-sepolia"
    test_transaction = TransactionRequestEIP1559(
        to="0x1234567890123456789012345678901234567890",
        value=w3.to_wei(0.000001, "ether"),
    )

    result = await server_account.send_transaction(
        network=test_network,
        transaction=test_transaction,
    )

    mock_api_instance.send_evm_transaction.assert_called_once_with(
        address=server_account.address,
        send_evm_transaction_request=SendEvmTransactionRequest(
            transaction="0x02e5808080808094123456789012345678901234567890123456789085e8d4a5100080c0808080",
            network=test_network,
        ),
        x_idempotency_key=None,
    )
    assert result == "0x456"


@pytest.mark.asyncio
@patch("cdp.evm_server_account.EVMAccountsApi")
async def test_send_transaction_dynamic_fee(mock_api, server_account_model_factory):
    """Test send_transaction method with dynamic fee transaction."""
    address = "0x1234567890123456789012345678901234567890"
    name = "test-account"
    server_account_model = server_account_model_factory(address, name)
    mock_api_instance = mock_api.return_value
    mock_api_instance.send_evm_transaction = AsyncMock(
        return_value=SendEvmTransaction200Response(transaction_hash="0x789")
    )
    server_account = EvmServerAccount(server_account_model, mock_api_instance, mock_api_instance)

    w3 = Web3()

    test_network = "base-sepolia"
    test_transaction = DynamicFeeTransaction.from_dict(
        {
            "to": "0x1234567890123456789012345678901234567890",
            "value": w3.to_wei(0.000001, "ether"),
            "gas": 21000,
            "maxFeePerGas": 1000000000000000000,
            "maxPriorityFeePerGas": 1000000000000000000,
            "nonce": 1,
            "type": "0x2",
        }
    )

    result = await server_account.send_transaction(
        network=test_network,
        transaction=test_transaction,
    )

    mock_api_instance.send_evm_transaction.assert_called_once_with(
        address=server_account.address,
        send_evm_transaction_request=SendEvmTransactionRequest(
            transaction="0x02f78001880de0b6b3a7640000880de0b6b3a764000082520894123456789012345678901234567890123456789085e8d4a5100080c0808080",
            network=test_network,
        ),
        x_idempotency_key=None,
    )
    assert result == "0x789"


@pytest.mark.asyncio
async def test_list_token_balances(server_account_model_factory, evm_token_balances_model_factory):
    """Test list_token_balances method."""
    address = "0x1234567890123456789012345678901234567890"
    name = "test-account"
    server_account_model = server_account_model_factory(address, name)

    mock_evm_token_balances_api = AsyncMock()
    mock_api_clients = AsyncMock()
    mock_api_clients.evm_token_balances = mock_evm_token_balances_api

    mock_token_balances = evm_token_balances_model_factory()

    mock_evm_token_balances_api.list_evm_token_balances = AsyncMock(
        return_value=mock_token_balances
    )

    expected_result = ListTokenBalancesResult(
        balances=[
            EvmTokenBalance(
                token=EvmToken(
                    contract_address="0x1234567890123456789012345678901234567890",
                    network="base-sepolia",
                    symbol="TEST",
                    name="Test Token",
                ),
                amount=EvmTokenAmount(amount=1000000000000000000, decimals=18),
            ),
        ],
        next_page_token="next-page-token",
    )

    server_account = EvmServerAccount(server_account_model, mock_api_clients, mock_api_clients)

    result = await server_account.list_token_balances(network="base-sepolia")

    mock_evm_token_balances_api.list_evm_token_balances.assert_called_once_with(
        address=address, network="base-sepolia", page_size=None, page_token=None
    )

    assert result == expected_result


@pytest.mark.asyncio
async def test_quote_fund_transfer_usdc(
    server_account_model_factory, payment_transfer_model_factory, payment_method_model_factory
):
    """Test quote_fund method."""
    address = "0x1234567890123456789012345678901234567890"
    name = "test-account"
    server_account_model = server_account_model_factory(address, name)
    payment_transfer = payment_transfer_model_factory()
    payment_method = payment_method_model_factory()

    mock_payments_api = AsyncMock()

    mock_api_clients = AsyncMock(spec=ApiClients)
    mock_api_clients.payments = mock_payments_api

    mock_payments_api.get_payment_methods = AsyncMock(return_value=[payment_method])

    mock_payments_api.create_payment_transfer_quote = AsyncMock(
        return_value=CreatePaymentTransferQuote201Response(transfer=payment_transfer)
    )

    server_account = EvmServerAccount(server_account_model, mock_api_clients, mock_api_clients)
    # 1 USDC
    result = await server_account.quote_fund(network="base", token="usdc", amount=1000000)

    mock_payments_api.get_payment_methods.assert_called_once()

    mock_payments_api.create_payment_transfer_quote.assert_called_once()
    call_args = mock_payments_api.create_payment_transfer_quote.call_args[1]
    assert call_args["create_payment_transfer_quote_request"].source_type == "payment_method"
    assert call_args["create_payment_transfer_quote_request"].target_type == "crypto_rail"
    assert call_args["create_payment_transfer_quote_request"].amount == "1"
    assert call_args["create_payment_transfer_quote_request"].currency == "usdc"

    assert isinstance(result, Quote)
    assert result.quote_id == payment_transfer.id
    assert result.network == "base"
    assert result.token == "usdc"
    assert result.fiat_amount == "1"
    assert result.fiat_currency == "usd"
    assert result.token_amount == "1"
    assert result.token == "usdc"
    assert result.fees == []


@pytest.mark.asyncio
async def test_quote_fund_transfer_eth(
    server_account_model_factory, payment_transfer_model_factory, payment_method_model_factory
):
    """Test quote_fund method."""
    address = "0x1234567890123456789012345678901234567890"
    name = "test-account"
    server_account_model = server_account_model_factory(address, name)
    payment_transfer = payment_transfer_model_factory(
        source_amount="1000", source_currency="usd", target_amount="1.1", target_currency="eth"
    )
    payment_method = payment_method_model_factory()

    mock_payments_api = AsyncMock()

    mock_api_clients = AsyncMock(spec=ApiClients)
    mock_api_clients.payments = mock_payments_api

    mock_payments_api.get_payment_methods = AsyncMock(return_value=[payment_method])

    mock_payments_api.create_payment_transfer_quote = AsyncMock(
        return_value=CreatePaymentTransferQuote201Response(transfer=payment_transfer)
    )

    server_account = EvmServerAccount(server_account_model, mock_api_clients, mock_api_clients)
    # 1.1 ETH
    result = await server_account.quote_fund(
        network="base", token="eth", amount=1100000000000000000
    )

    mock_payments_api.get_payment_methods.assert_called_once()

    mock_payments_api.create_payment_transfer_quote.assert_called_once()
    call_args = mock_payments_api.create_payment_transfer_quote.call_args[1]
    assert call_args["create_payment_transfer_quote_request"].source_type == "payment_method"
    assert call_args["create_payment_transfer_quote_request"].target_type == "crypto_rail"
    assert call_args["create_payment_transfer_quote_request"].amount == "1.1"
    assert call_args["create_payment_transfer_quote_request"].currency == "eth"

    assert isinstance(result, Quote)
    assert result.quote_id == payment_transfer.id
    assert result.network == "base"
    assert result.token == "eth"
    assert result.fiat_amount == "1000"
    assert result.fiat_currency == "usd"
    assert result.token_amount == "1.1"
    assert result.token == "eth"
    assert result.fees == []


@pytest.mark.asyncio
async def test_fund_transfer_eth(
    server_account_model_factory, payment_transfer_model_factory, payment_method_model_factory
):
    """Test fund method."""
    address = "0x1234567890123456789012345678901234567890"
    name = "test-account"
    server_account_model = server_account_model_factory(address, name)
    payment_transfer = payment_transfer_model_factory(
        source_amount="1000", source_currency="usd", target_amount="1.1", target_currency="eth"
    )
    payment_method = payment_method_model_factory()

    mock_payments_api = AsyncMock()

    mock_api_clients = AsyncMock(spec=ApiClients)
    mock_api_clients.payments = mock_payments_api

    mock_payments_api.get_payment_methods = AsyncMock(return_value=[payment_method])

    mock_payments_api.create_payment_transfer_quote = AsyncMock(
        return_value=CreatePaymentTransferQuote201Response(transfer=payment_transfer)
    )

    server_account = EvmServerAccount(server_account_model, mock_api_clients, mock_api_clients)
    # 1.1 ETH
    result = await server_account.fund(network="base", token="eth", amount=1100000000000000000)

    mock_payments_api.get_payment_methods.assert_called_once()

    mock_payments_api.create_payment_transfer_quote.assert_called_once()
    call_args = mock_payments_api.create_payment_transfer_quote.call_args[1]
    assert call_args["create_payment_transfer_quote_request"].source_type == "payment_method"
    assert call_args["create_payment_transfer_quote_request"].target_type == "crypto_rail"
    assert call_args["create_payment_transfer_quote_request"].amount == "1.1"
    assert call_args["create_payment_transfer_quote_request"].currency == "eth"
    assert call_args["create_payment_transfer_quote_request"].execute is True

    assert result.id == payment_transfer.id
    assert result.network == payment_transfer.target.actual_instance.network
    assert result.target_amount == payment_transfer.target_amount
    assert result.target_currency == payment_transfer.target_currency
    assert result.status == payment_transfer.status
    assert result.transaction_hash == payment_transfer.transaction_hash


@pytest.mark.asyncio
async def test_fund_transfer_usdc(
    server_account_model_factory, payment_transfer_model_factory, payment_method_model_factory
):
    """Test fund method with USDC."""
    address = "0x1234567890123456789012345678901234567890"
    name = "test-account"
    server_account_model = server_account_model_factory(address, name)
    payment_transfer = payment_transfer_model_factory(
        source_amount="1", source_currency="usd", target_amount="1", target_currency="usdc"
    )
    payment_method = payment_method_model_factory()

    mock_payments_api = AsyncMock()

    mock_api_clients = AsyncMock(spec=ApiClients)
    mock_api_clients.payments = mock_payments_api

    mock_payments_api.get_payment_methods = AsyncMock(return_value=[payment_method])

    mock_payments_api.create_payment_transfer_quote = AsyncMock(
        return_value=CreatePaymentTransferQuote201Response(transfer=payment_transfer)
    )

    server_account = EvmServerAccount(server_account_model, mock_api_clients, mock_api_clients)
    # 1 USDC (6 decimals)
    result = await server_account.fund(network="base", token="usdc", amount=1000000)

    mock_payments_api.get_payment_methods.assert_called_once()

    mock_payments_api.create_payment_transfer_quote.assert_called_once()
    call_args = mock_payments_api.create_payment_transfer_quote.call_args[1]
    assert call_args["create_payment_transfer_quote_request"].source_type == "payment_method"
    assert call_args["create_payment_transfer_quote_request"].target_type == "crypto_rail"
    assert call_args["create_payment_transfer_quote_request"].amount == "1"
    assert call_args["create_payment_transfer_quote_request"].currency == "usdc"
    assert call_args["create_payment_transfer_quote_request"].execute is True

    assert result.id == payment_transfer.id
    assert result.network == payment_transfer.target.actual_instance.network
    assert result.target_amount == payment_transfer.target_amount
    assert result.target_currency == payment_transfer.target_currency
    assert result.status == payment_transfer.status
    assert result.transaction_hash == payment_transfer.transaction_hash


@pytest.mark.asyncio
async def test_wait_for_fund_operation_receipt_success(
    server_account_model_factory, payment_transfer_model_factory
):
    """Test wait_for_fund_operation_receipt method with successful completion."""
    address = "0x1234567890123456789012345678901234567890"
    name = "test-account"
    server_account_model = server_account_model_factory(address, name)

    mock_payments_api = AsyncMock()
    mock_api_clients = AsyncMock(spec=ApiClients)
    mock_api_clients.payments = mock_payments_api

    completed_transfer = payment_transfer_model_factory(status="completed")
    mock_payments_api.get_payment_transfer = AsyncMock(return_value=completed_transfer)

    server_account = EvmServerAccount(server_account_model, mock_api_clients, mock_api_clients)
    result = await server_account.wait_for_fund_operation_receipt(transfer_id="test-transfer-id")

    assert result == completed_transfer
    mock_payments_api.get_payment_transfer.assert_called_once_with("test-transfer-id")


@pytest.mark.asyncio
async def test_wait_for_fund_operation_receipt_failure(
    server_account_model_factory, payment_transfer_model_factory
):
    """Test wait_for_fund_operation_receipt method with failed transfer."""
    address = "0x1234567890123456789012345678901234567890"
    name = "test-account"
    server_account_model = server_account_model_factory(address, name)

    mock_payments_api = AsyncMock()
    mock_api_clients = AsyncMock(spec=ApiClients)
    mock_api_clients.payments = mock_payments_api

    failed_transfer = payment_transfer_model_factory(status="failed")
    mock_payments_api.get_payment_transfer = AsyncMock(return_value=failed_transfer)

    server_account = EvmServerAccount(server_account_model, mock_api_clients, mock_api_clients)
    result = await server_account.wait_for_fund_operation_receipt(transfer_id="test-transfer-id")

    assert result == failed_transfer
    mock_payments_api.get_payment_transfer.assert_called_once_with("test-transfer-id")


@pytest.mark.asyncio
async def test_wait_for_fund_operation_receipt_timeout(
    server_account_model_factory, payment_transfer_model_factory
):
    """Test wait_for_fund_operation_receipt method with timeout."""
    address = "0x1234567890123456789012345678901234567890"
    name = "test-account"
    server_account_model = server_account_model_factory(address, name)

    mock_payments_api = AsyncMock()
    mock_api_clients = AsyncMock(spec=ApiClients)
    mock_api_clients.payments = mock_payments_api

    pending_transfer = payment_transfer_model_factory(status="pending")
    mock_payments_api.get_payment_transfer = AsyncMock(return_value=pending_transfer)

    server_account = EvmServerAccount(server_account_model, mock_api_clients, mock_api_clients)

    with pytest.raises(TimeoutError):
        await server_account.wait_for_fund_operation_receipt(
            transfer_id="test-transfer-id", timeout_seconds=0.1, interval_seconds=0.1
        )



# coding: utf-8

"""
    Coinbase Developer Platform APIs

    The Coinbase Developer Platform APIs - leading the world's transition onchain.

    The version of the OpenAPI document: 2.0.0
    Contact: cdp@coinbase.com
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, Field, StrictStr
from typing import Any, ClassVar, Dict, List
from cdp.openapi_client.models.eip712_domain import EIP712Domain
from typing import Optional, Set
from typing_extensions import Self

class EIP712Message(BaseModel):
    """
    The message to sign using EIP-712.
    """ # noqa: E501
    domain: EIP712Domain
    types: Dict[str, Any] = Field(description="A mapping of struct names to an array of type objects (name + type). Each key corresponds to a type name (e.g., \"`EIP712Domain`\", \"`PermitTransferFrom`\"). ")
    primary_type: StrictStr = Field(description="The primary type of the message. This is the name of the struct in the `types` object that is the root of the message.", alias="primaryType")
    message: Dict[str, Any] = Field(description="The message to sign. The structure of this message must match the `primaryType` struct in the `types` object.")
    __properties: ClassVar[List[str]] = ["domain", "types", "primaryType", "message"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of EIP712Message from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of domain
        if self.domain:
            _dict['domain'] = self.domain.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of EIP712Message from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "domain": EIP712Domain.from_dict(obj["domain"]) if obj.get("domain") is not None else None,
            "types": obj.get("types"),
            "primaryType": obj.get("primaryType"),
            "message": obj.get("message")
        })
        return _obj





from cdp.openapi_client.api.solana_accounts_api import SolanaAccountsApi
from cdp.openapi_client.models.sign_solana_message200_response import (
    SignSolanaMessage200Response as SignSolanaMessageResponse,
)
from cdp.openapi_client.models.sign_solana_message_request import SignSolanaMessageRequest


async def sign_message(
    solana_accounts_api: SolanaAccountsApi,
    address: str,
    message: str,
    idempotency_key: str,
) -> SignSolanaMessageResponse:
    """Sign a message.

    Args:
        solana_accounts_api (SolanaAccountsApi): The Solana accounts API.
        address (str): The address of the Solana account.
        message (str): The message to sign.
        idempotency_key (str): The idempotency key.

    Returns:
        SignSolanaMessageResponse: The signature of the message.

    """
    return await solana_accounts_api.sign_solana_message(
        sign_solana_message_request=SignSolanaMessageRequest(
            message=message,
        ),
        address=address,
        x_idempotency_key=idempotency_key,
    )



# coding: utf-8

"""
    Coinbase Developer Platform APIs

    The Coinbase Developer Platform APIs - leading the world's transition onchain.

    The version of the OpenAPI document: 2.0.0
    Contact: cdp@coinbase.com
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, Field, field_validator
from typing import Any, ClassVar, Dict, List, Optional
from typing_extensions import Annotated
from cdp.openapi_client.models.evm_swaps_network import EvmSwapsNetwork
from typing import Optional, Set
from typing_extensions import Self

class CreateEvmSwapQuoteRequest(BaseModel):
    """
    CreateEvmSwapQuoteRequest
    """ # noqa: E501
    network: EvmSwapsNetwork
    to_token: Annotated[str, Field(strict=True)] = Field(description="The 0x-prefixed contract address of the token to receive.", alias="toToken")
    from_token: Annotated[str, Field(strict=True)] = Field(description="The 0x-prefixed contract address of the token to send.", alias="fromToken")
    from_amount: Annotated[str, Field(strict=True)] = Field(description="The amount of the `fromToken` to send in atomic units of the token. For example, `1000000000000000000` when sending ETH equates to 1 ETH, `1000000` when sending USDC equates to 1 USDC, etc.", alias="fromAmount")
    taker: Annotated[str, Field(strict=True)] = Field(description="The 0x-prefixed address that holds the `fromToken` balance and has the `Permit2` allowance set for the swap.")
    signer_address: Optional[Annotated[str, Field(strict=True)]] = Field(default=None, description="The 0x-prefixed Externally Owned Account (EOA) address that will sign the `Permit2` EIP-712 permit message. This is only needed if `taker` is a smart contract.", alias="signerAddress")
    gas_price: Optional[Annotated[str, Field(strict=True)]] = Field(default=None, description="The target gas price for the swap transaction, in Wei. For EIP-1559 transactions, this value should be seen as the `maxFeePerGas` value. If not provided, the API will use an estimate based on the current network conditions.", alias="gasPrice")
    slippage_bps: Optional[Annotated[int, Field(le=10000, strict=True, ge=0)]] = Field(default=100, description="The maximum acceptable slippage of the `toToken` in basis points. If this parameter is set to 0, no slippage will be tolerated. If not provided, the default slippage tolerance is 100 bps (i.e., 1%).", alias="slippageBps")
    __properties: ClassVar[List[str]] = ["network", "toToken", "fromToken", "fromAmount", "taker", "signerAddress", "gasPrice", "slippageBps"]

    @field_validator('to_token')
    def to_token_validate_regular_expression(cls, value):
        """Validates the regular expression"""
        if not re.match(r"^0x[a-fA-F0-9]{40}$", value):
            raise ValueError(r"must validate the regular expression /^0x[a-fA-F0-9]{40}$/")
        return value

    @field_validator('from_token')
    def from_token_validate_regular_expression(cls, value):
        """Validates the regular expression"""
        if not re.match(r"^0x[a-fA-F0-9]{40}$", value):
            raise ValueError(r"must validate the regular expression /^0x[a-fA-F0-9]{40}$/")
        return value

    @field_validator('from_amount')
    def from_amount_validate_regular_expression(cls, value):
        """Validates the regular expression"""
        if not re.match(r"^\d+$", value):
            raise ValueError(r"must validate the regular expression /^\d+$/")
        return value

    @field_validator('taker')
    def taker_validate_regular_expression(cls, value):
        """Validates the regular expression"""
        if not re.match(r"^0x[a-fA-F0-9]{40}$", value):
            raise ValueError(r"must validate the regular expression /^0x[a-fA-F0-9]{40}$/")
        return value

    @field_validator('signer_address')
    def signer_address_validate_regular_expression(cls, value):
        """Validates the regular expression"""
        if value is None:
            return value

        if not re.match(r"^0x[a-fA-F0-9]{40}$", value):
            raise ValueError(r"must validate the regular expression /^0x[a-fA-F0-9]{40}$/")
        return value

    @field_validator('gas_price')
    def gas_price_validate_regular_expression(cls, value):
        """Validates the regular expression"""
        if value is None:
            return value

        if not re.match(r"^\d+$", value):
            raise ValueError(r"must validate the regular expression /^\d+$/")
        return value

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of CreateEvmSwapQuoteRequest from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of CreateEvmSwapQuoteRequest from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "network": obj.get("network"),
            "toToken": obj.get("toToken"),
            "fromToken": obj.get("fromToken"),
            "fromAmount": obj.get("fromAmount"),
            "taker": obj.get("taker"),
            "signerAddress": obj.get("signerAddress"),
            "gasPrice": obj.get("gasPrice"),
            "slippageBps": obj.get("slippageBps") if obj.get("slippageBps") is not None else 100
        })
        return _obj





# coding: utf-8

"""
    Coinbase Developer Platform APIs

    The Coinbase Developer Platform APIs - leading the world's transition onchain.

    The version of the OpenAPI document: 2.0.0
    Contact: cdp@coinbase.com
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


import unittest

from cdp.openapi_client.models.create_swap_quote_response_all_of_permit2 import CreateSwapQuoteResponseAllOfPermit2

class TestCreateSwapQuoteResponseAllOfPermit2(unittest.TestCase):
    """CreateSwapQuoteResponseAllOfPermit2 unit test stubs"""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def make_instance(self, include_optional) -> CreateSwapQuoteResponseAllOfPermit2:
        """Test CreateSwapQuoteResponseAllOfPermit2
            include_optional is a boolean, when False only required
            params are included, when True both required and
            optional params are included """
        # uncomment below to create an instance of `CreateSwapQuoteResponseAllOfPermit2`
        """
        model = CreateSwapQuoteResponseAllOfPermit2()
        if include_optional:
            return CreateSwapQuoteResponseAllOfPermit2(
                hash = '0x1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef',
                eip712 = {"domain":{"name":"Permit2","chainId":1,"verifyingContract":"0x000000000022D473030F116dDEE9F6B43aC78BA3"},"types":{"EIP712Domain":[{"name":"name","type":"string"},{"name":"chainId","type":"uint256"},{"name":"verifyingContract","type":"address"}],"PermitTransferFrom":[{"name":"permitted","type":"TokenPermissions"},{"name":"spender","type":"address"},{"name":"nonce","type":"uint256"},{"name":"deadline","type":"uint256"}],"TokenPermissions":[{"name":"token","type":"address"},{"name":"amount","type":"uint256"}]},"primaryType":"PermitTransferFrom","message":{"permitted":{"token":"0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48","amount":"1000000"},"spender":"0xFfFfFfFFfFFfFFfFFfFFFFFffFFFffffFfFFFfFf","nonce":"123456","deadline":"1717123200"}}
            )
        else:
            return CreateSwapQuoteResponseAllOfPermit2(
                hash = '0x1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef',
                eip712 = {"domain":{"name":"Permit2","chainId":1,"verifyingContract":"0x000000000022D473030F116dDEE9F6B43aC78BA3"},"types":{"EIP712Domain":[{"name":"name","type":"string"},{"name":"chainId","type":"uint256"},{"name":"verifyingContract","type":"address"}],"PermitTransferFrom":[{"name":"permitted","type":"TokenPermissions"},{"name":"spender","type":"address"},{"name":"nonce","type":"uint256"},{"name":"deadline","type":"uint256"}],"TokenPermissions":[{"name":"token","type":"address"},{"name":"amount","type":"uint256"}]},"primaryType":"PermitTransferFrom","message":{"permitted":{"token":"0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48","amount":"1000000"},"spender":"0xFfFfFfFFfFFfFFfFFfFFFFFffFFFffffFfFFFfFf","nonce":"123456","deadline":"1717123200"}},
        )
        """

    def testCreateSwapQuoteResponseAllOfPermit2(self):
        """Test CreateSwapQuoteResponseAllOfPermit2"""
        # inst_req_only = self.make_instance(include_optional=False)
        # inst_req_and_optional = self.make_instance(include_optional=True)

if __name__ == '__main__':
    unittest.main()



from unittest.mock import MagicMock, patch
from urllib.parse import urlparse

from cdp.auth.clients.urllib3.client import Urllib3AuthClient
from cdp.auth.utils.http import GetAuthHeadersOptions


@patch("cdp.auth.utils.http.generate_jwt", return_value="mock.jwt.token")
@patch("urllib3.PoolManager")
@patch("cdp.auth.clients.urllib3.client.get_auth_headers")
def test_request_adds_auth_headers(
    mock_get_headers, mock_pool_manager, mock_jwt, auth_client_options_factory
):
    """Test that requests include authentication headers."""
    # Setup
    options = auth_client_options_factory()
    base_url = "https://api.example.com"

    mock_client = MagicMock()
    mock_client.request = MagicMock()
    mock_pool_manager.return_value = mock_client

    mock_get_headers.return_value = {
        "Authorization": "Bearer test.token",
        "Content-Type": "application/json",
    }

    client = Urllib3AuthClient(options, base_url)

    # Execute
    client.request("GET", "/test")

    # Verify
    mock_get_headers.assert_called_once()
    call_args = mock_get_headers.call_args[0][0]
    assert isinstance(call_args, GetAuthHeadersOptions)
    assert call_args.api_key_id == "test-key-id"
    assert call_args.request_method == "GET"

    parsed_url = urlparse(f"https://{call_args.request_host}")
    assert parsed_url.hostname == "api.example.com"


@patch("cdp.auth.utils.http.generate_jwt", return_value="mock.jwt.token")
@patch("urllib3.PoolManager")
@patch("cdp.auth.clients.urllib3.client.get_auth_headers")
@patch("cdp.auth.clients.urllib3.client.logger")
def test_request_with_debug(
    mock_logger,
    mock_get_headers,
    mock_pool_manager,
    mock_jwt,
    auth_client_options_factory,
    http_client_factory,
    http_response_factory,
    auth_headers_factory,
):
    """Test client with debug logging enabled."""
    # Setup
    options = auth_client_options_factory()
    base_url = "https://api.example.com"

    mock_response = http_response_factory()
    mock_client = http_client_factory(response=mock_response)
    mock_pool_manager.return_value = mock_client

    mock_get_headers.return_value = auth_headers_factory()

    # Execute
    client = Urllib3AuthClient(options, base_url, debug=True)
    response = client.request("GET", "/test")

    # Verify
    assert mock_logger.debug.call_count == 2

    request_log_call = mock_logger.debug.call_args_list[0]
    log_format, method, url, headers, body = request_log_call[0]
    assert "HTTP Request:" in log_format
    assert method == "GET"
    assert "/test" in url

    response_log_call = mock_logger.debug.call_args_list[1]
    log_format, status, headers, body = response_log_call[0]
    assert "HTTP Response:" in log_format
    assert status == 200
    assert response.data == b'{"test": "data"}'

    assert response is mock_response
    assert response.status == 200
    assert response.data == b'{"test": "data"}'
    assert response.headers == {"Content-Type": "application/json"}


@patch("cdp.auth.utils.http.generate_jwt", return_value="mock.jwt.token")
@patch("urllib3.PoolManager")
@patch("cdp.auth.clients.urllib3.client.get_auth_headers")
def test_request_url_handling(
    mock_get_headers, mock_pool_manager, mock_jwt, auth_client_options_factory
):
    """Test handling of relative and absolute URLs."""
    # Setup
    options = auth_client_options_factory()
    base_url = "https://api.example.com"

    mock_client = MagicMock()
    mock_client.request = MagicMock()
    mock_pool_manager.return_value = mock_client

    mock_get_headers.return_value = {
        "Authorization": "Bearer test.token",
        "Content-Type": "application/json",
    }

    client = Urllib3AuthClient(options, base_url)

    # Test with relative URL
    client.request("GET", "/test")
    called_url = mock_client.request.call_args[1]["url"]
    assert called_url.startswith(base_url)

    # Test with absolute URL
    absolute_url = "https://other.example.com/test"
    client.request("GET", absolute_url)
    called_url = mock_client.request.call_args[1]["url"]
    assert called_url == absolute_url


@patch("cdp.auth.utils.http.generate_jwt", return_value="mock.jwt.token")
@patch("urllib3.PoolManager")
@patch("cdp.auth.clients.urllib3.client.get_auth_headers")
def test_request_merges_headers(
    mock_get_headers, mock_pool_manager, mock_jwt, auth_client_options_factory
):
    """Test that custom headers are merged with auth headers."""
    # Setup
    options = auth_client_options_factory()
    base_url = "https://api.example.com"

    mock_client = MagicMock()
    mock_client.request = MagicMock()
    mock_pool_manager.return_value = mock_client

    mock_get_headers.return_value = {
        "Authorization": "Bearer test.token",
        "Content-Type": "application/json",
    }

    client = Urllib3AuthClient(options, base_url)

    # Execute
    custom_headers = {"Custom-Header": "value"}
    client.request("GET", "/test", headers=custom_headers)

    # Verify
    actual_headers = mock_client.request.call_args[1]["headers"]
    assert "Custom-Header" in actual_headers
    assert "Authorization" in actual_headers


@patch("cdp.auth.utils.http.generate_jwt", return_value="mock.jwt.token")
@patch("urllib3.PoolManager")
@patch("cdp.auth.clients.urllib3.client.get_auth_headers")
def test_request_with_query_params(
    mock_get_headers, mock_pool_manager, mock_jwt, auth_client_options_factory
):
    """Test handling of query parameters in the URL."""
    # Setup
    options = auth_client_options_factory()
    base_url = "https://api.example.com"

    mock_client = MagicMock()
    mock_client.request = MagicMock()
    mock_pool_manager.return_value = mock_client

    mock_get_headers.return_value = {
        "Authorization": "Bearer test.token",
        "Content-Type": "application/json",
    }

    client = Urllib3AuthClient(options, base_url)

    # Execute
    client.request("GET", "/test", fields={"key": "value"})

    # Verify
    call_kwargs = mock_client.request.call_args[1]
    assert "fields" in call_kwargs
    assert call_kwargs["fields"] == {"key": "value"}
    assert call_kwargs["url"] == f"{base_url}/test"


@patch("cdp.auth.utils.http.generate_jwt", return_value="mock.jwt.token")
@patch("urllib3.PoolManager")
@patch("cdp.auth.clients.urllib3.client.get_auth_headers")
def test_request_with_json_body(
    mock_get_headers, mock_pool_manager, mock_jwt, auth_client_options_factory
):
    """Test request with JSON body."""
    # Setup
    options = auth_client_options_factory()
    base_url = "https://api.example.com"

    mock_client = MagicMock()
    mock_client.request = MagicMock()
    mock_pool_manager.return_value = mock_client

    mock_get_headers.return_value = {
        "Authorization": "Bearer test.token",
        "Content-Type": "application/json",
    }

    client = Urllib3AuthClient(options, base_url)

    # Execute
    test_body = {"data": "test"}
    client.request("POST", "/test", body=test_body)

    # Verify
    actual_headers = mock_client.request.call_args[1]["headers"]
    assert actual_headers.get("Content-Type") == "application/json"



from cdp.openapi_client.models.eth_value_criterion import EthValueCriterion
from cdp.openapi_client.models.evm_address_criterion import EvmAddressCriterion
from cdp.openapi_client.models.evm_message_criterion import EvmMessageCriterion
from cdp.openapi_client.models.evm_network_criterion import EvmNetworkCriterion
from cdp.openapi_client.models.rule import Rule
from cdp.openapi_client.models.send_evm_transaction_criteria_inner import (
    SendEvmTransactionCriteriaInner,
)
from cdp.openapi_client.models.send_evm_transaction_rule import SendEvmTransactionRule
from cdp.openapi_client.models.sign_evm_hash_rule import SignEvmHashRule
from cdp.openapi_client.models.sign_evm_message_criteria_inner import SignEvmMessageCriteriaInner
from cdp.openapi_client.models.sign_evm_message_rule import SignEvmMessageRule
from cdp.openapi_client.models.sign_evm_transaction_criteria_inner import (
    SignEvmTransactionCriteriaInner,
)
from cdp.openapi_client.models.sign_evm_transaction_rule import SignEvmTransactionRule
from cdp.openapi_client.models.sign_sol_transaction_criteria_inner import (
    SignSolTransactionCriteriaInner,
)
from cdp.openapi_client.models.sign_sol_transaction_rule import SignSolTransactionRule
from cdp.openapi_client.models.sol_address_criterion import SolAddressCriterion
from cdp.policies.types import Rule as RuleType

# OpenAPI criterion constructor mapping per operation
openapi_criterion_mapping = {
    "sendEvmTransaction": {
        "ethValue": lambda c: SendEvmTransactionCriteriaInner(
            actual_instance=EthValueCriterion(
                eth_value=c.ethValue,
                operator=c.operator,
                type="ethValue",
            )
        ),
        "evmAddress": lambda c: SendEvmTransactionCriteriaInner(
            actual_instance=EvmAddressCriterion(
                addresses=c.addresses,
                operator=c.operator,
                type="evmAddress",
            )
        ),
        "evmNetwork": lambda c: SendEvmTransactionCriteriaInner(
            actual_instance=EvmNetworkCriterion(
                networks=c.networks,
                operator=c.operator,
                type="evmNetwork",
            )
        ),
    },
    "signEvmTransaction": {
        "ethValue": lambda c: SignEvmTransactionCriteriaInner(
            actual_instance=EthValueCriterion(
                eth_value=c.ethValue,
                operator=c.operator,
                type="ethValue",
            )
        ),
        "evmAddress": lambda c: SignEvmTransactionCriteriaInner(
            actual_instance=EvmAddressCriterion(
                addresses=c.addresses,
                operator=c.operator,
                type="evmAddress",
            )
        ),
    },
    "signEvmHash": {},
    "signEvmMessage": {
        "evmMessage": lambda c: SignEvmMessageCriteriaInner(
            actual_instance=EvmMessageCriterion(
                match=c.match,
                type="evmMessage",
            )
        ),
    },
    "signSolTransaction": {
        "solAddress": lambda c: SignSolTransactionCriteriaInner(
            actual_instance=SolAddressCriterion(
                addresses=c.addresses,
                operator=c.operator,
                type="solAddress",
            )
        ),
    },
}

# OpenAPI rule constructor mapping
openapi_rule_mapping = {
    "sendEvmTransaction": SendEvmTransactionRule,
    "signEvmTransaction": SignEvmTransactionRule,
    "signEvmHash": SignEvmHashRule,
    "signEvmMessage": SignEvmMessageRule,
    "signSolTransaction": SignSolTransactionRule,
}


def map_request_rules_to_openapi_format(request_rules: list[RuleType]) -> list[Rule]:
    """Build a properly formatted list of OpenAPI policy rules from a list of request rules.

    Args:
        request_rules (List[RuleType]): The request rules to build from.

    Returns:
        List[Rule]: A list of rules formatted for the OpenAPI policy.

    """
    rules = []
    for rule in request_rules:
        if rule.operation not in openapi_criterion_mapping:
            raise ValueError(f"Unknown operation {rule.operation}")

        rule_cls = openapi_rule_mapping[rule.operation]

        if not hasattr(rule, "criteria"):
            rules.append(
                Rule(
                    actual_instance=rule_cls(
                        action=rule.action,
                        operation=rule.operation,
                    )
                )
            )
            continue

        criteria_builders = openapi_criterion_mapping[rule.operation]
        criteria = []

        for criterion in rule.criteria:
            if criterion.type not in criteria_builders:
                raise ValueError(
                    f"Unknown criterion type {criterion.type} for operation {rule.operation}"
                )
            criteria.append(criteria_builders[criterion.type](criterion))

        rules.append(
            Rule(
                actual_instance=rule_cls(
                    action=rule.action,
                    operation=rule.operation,
                    criteria=criteria,
                )
            )
        )

    return rules



from typing import Any, Literal

from eth_account.datastructures import (
    SignedMessage,
    SignedTransaction,
)
from eth_account.messages import (
    SignableMessage,
    _hash_eip191_message,
)
from eth_account.signers.base import BaseAccount
from eth_account.typed_transactions import DynamicFeeTransaction, TypedTransaction
from eth_account.types import (
    TransactionDictType,
)
from eth_typing import (
    Hash32,
)
from hexbytes import HexBytes
from pydantic import BaseModel, ConfigDict, Field
from web3 import Web3

from cdp.actions.evm.fund import (
    FundOptions,
    QuoteFundOptions,
    fund,
    quote_fund,
    wait_for_fund_operation_receipt,
)
from cdp.actions.evm.fund.quote import Quote
from cdp.actions.evm.fund.types import FundOperationResult
from cdp.actions.evm.list_token_balances import list_token_balances
from cdp.actions.evm.request_faucet import request_faucet
from cdp.actions.evm.send_transaction import send_transaction
from cdp.actions.evm.swap import SwapOptions
from cdp.actions.evm.swap.types import QuoteSwapResult, SwapResult
from cdp.api_clients import ApiClients
from cdp.evm_token_balances import ListTokenBalancesResult
from cdp.evm_transaction_types import TransactionRequestEIP1559
from cdp.openapi_client.api.evm_accounts_api import EVMAccountsApi
from cdp.openapi_client.models.eip712_domain import EIP712Domain
from cdp.openapi_client.models.eip712_message import EIP712Message
from cdp.openapi_client.models.evm_account import EvmAccount as EvmServerAccountModel
from cdp.openapi_client.models.sign_evm_hash_request import SignEvmHashRequest
from cdp.openapi_client.models.sign_evm_message_request import SignEvmMessageRequest
from cdp.openapi_client.models.sign_evm_transaction_request import (
    SignEvmTransactionRequest,
)
from cdp.openapi_client.models.transfer import Transfer


class EvmServerAccount(BaseAccount, BaseModel):
    """A class representing an EVM server account."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __init__(
        self,
        evm_server_account_model: EvmServerAccountModel,
        evm_accounts_api: EVMAccountsApi,
        api_clients: ApiClients,
    ) -> None:
        """Initialize the EvmServerAccount class.

        Args:
            evm_server_account_model (EvmServerAccountModel): The EVM server account model.
            evm_accounts_api (EVMAccountsApi): The EVM accounts API.
            api_clients (ApiClients): The API client.

        """
        super().__init__()

        self.__address = evm_server_account_model.address
        self.__name = evm_server_account_model.name
        self.__policies = evm_server_account_model.policies
        self.__evm_accounts_api = evm_accounts_api
        self.__api_clients = api_clients

    @property
    def address(self) -> str:
        """Get the EVM Account Address.

        Returns:
            str: The EVM account address.

        """
        return self.__address

    @property
    def name(self) -> str | None:
        """Get the name of the EVM account.

        Returns:
            str | None: The name of the EVM account.

        """
        return self.__name

    @property
    def policies(self) -> list[str]:
        """Gets the list of policies the apply to this account.

        Returns:
            str: The list of Policy IDs.

        """
        return self.__policies

    async def sign_message(
        self, signable_message: SignableMessage, idempotency_key: str | None = None
    ) -> SignedMessage:
        """Sign the EIP-191 message.

        Args:
            signable_message: The encoded message, ready for signing
            idempotency_key: Optional idempotency key

        Returns:
            The signed message

        Raises:
            AttributeError: If the signature response is missing required fields

        """
        message_body = signable_message.body
        message_hex = (
            message_body.hex() if isinstance(message_body, bytes) else HexBytes(message_body).hex()
        )
        signature_response = await self.__evm_accounts_api.sign_evm_message(
            address=self.address,
            sign_evm_message_request=SignEvmMessageRequest(message=message_hex),
            x_idempotency_key=idempotency_key,
        )

        message_hash = _hash_eip191_message(signable_message)

        signature_bytes = HexBytes(signature_response.signature)
        r = int.from_bytes(signature_bytes[0:32], byteorder="big")
        s = int.from_bytes(signature_bytes[32:64], byteorder="big")
        v = signature_bytes[64]

        return SignedMessage(
            message_hash=message_hash,
            r=r,
            s=s,
            v=v,
            signature=signature_bytes,
        )

    async def unsafe_sign_hash(
        self, message_hash: Hash32, idempotency_key: str | None = None
    ) -> SignedMessage:
        """Sign the hash of a message.

        WARNING: Never sign a hash that you didn't generate,
        it can be an arbitrary transaction.

        Args:
            message_hash: 32 byte hash of the message to sign
            idempotency_key: Optional idempotency key

        Returns:
            The signed message

        Raises:
            ValueError: If the signature response is missing required fields

        """
        hash_hex = HexBytes(message_hash).hex()
        sign_evm_hash_request = SignEvmHashRequest(hash=hash_hex)
        signature_response = await self.__evm_accounts_api.sign_evm_hash(
            address=self.address,
            sign_evm_hash_request=sign_evm_hash_request,
            x_idempotency_key=idempotency_key,
        )

        signature_bytes = HexBytes(signature_response.signature)
        r = int.from_bytes(signature_bytes[0:32], byteorder="big")
        s = int.from_bytes(signature_bytes[32:64], byteorder="big")
        v = signature_bytes[64]

        return SignedMessage(
            message_hash=message_hash,
            r=r,
            s=s,
            v=v,
            signature=signature_bytes,
        )

    async def sign_transaction(
        self, transaction_dict: TransactionDictType, idempotency_key: str | None = None
    ) -> SignedTransaction:
        """Sign a transaction dict.

        Args:
            transaction_dict: transaction with all fields specified
            idempotency_key: Optional idempotency key
        Returns:
            The signed transaction
        Raises:
            ValueError: If the signature response is missing required fields

        """
        typed_tx = TypedTransaction.from_dict(transaction_dict)
        typed_tx.transaction.dictionary["v"] = 0
        typed_tx.transaction.dictionary["r"] = 0
        typed_tx.transaction.dictionary["s"] = 0
        payload = typed_tx.transaction.payload()
        serialized_tx = bytes([typed_tx.transaction_type]) + payload

        sign_evm_transaction_request = SignEvmTransactionRequest(
            transaction="0x" + serialized_tx.hex()
        )
        signature_response = await self.__evm_accounts_api.sign_evm_transaction(
            address=self.address,
            sign_evm_transaction_request=sign_evm_transaction_request,
            x_idempotency_key=idempotency_key,
        )

        # Get the signed transaction bytes
        signed_tx_bytes = HexBytes(signature_response.signed_transaction)
        transaction_hash = Web3.keccak(signed_tx_bytes)

        # Extract signature components from the response
        signature_bytes = HexBytes(signature_response.signed_transaction)
        r = int.from_bytes(signature_bytes[0:32], byteorder="big")
        s = int.from_bytes(signature_bytes[32:64], byteorder="big")
        v = signature_bytes[64]

        return SignedTransaction(
            raw_transaction=signed_tx_bytes,
            hash=transaction_hash,
            r=r,
            s=s,
            v=v,
        )

    async def transfer(self, to: str | BaseAccount, amount: int, token: str, network: str):
        """Transfer an amount of a token from an account to another account.

        Args:
            to: The account or 0x-prefixed address to transfer the token to.
            amount: The amount of the token to transfer, represented as an atomic unit (e.g. 10000 for 0.01 USDC).
            The cdp module exports a `parse_units` util to convert to atomic units.
            Otherwise, you can pass atomic units directly. See examples below.
            token: The token to transfer.
            network: The network to transfer the token on.

        Returns:
            The result of the transfer.

        Examples:
            >>> transfer = await sender.transfer(
            ...     to="0x9F663335Cd6Ad02a37B633602E98866CF944124d",
            ...     amount=10000,  # equivalent to 0.01 USDC
            ...     token="usdc",
            ...     network="base-sepolia",
            ... )

            **Using parse_units to specify USDC amount**
            >>> from cdp import parse_units
            >>> transfer = await sender.transfer(
            ...     to="0x9F663335Cd6Ad02a37B633602E98866CF944124d",
            ...     amount=parse_units("0.01", 6),  # USDC uses 6 decimal places
            ...     token="usdc",
            ...     network="base-sepolia",
            ... )

            **Transfer to another account**
            >>> sender = await cdp.evm.create_account(name="Sender")
            >>> receiver = await cdp.evm.create_account(name="Receiver")
            >>>
            >>> transfer = await sender.transfer({
            ...     "to": receiver,
            ...     "amount": 10000,  # equivalent to 0.01 USDC
            ...     "token": "usdc",
            ...     "network": "base-sepolia",
            ... })

        """
        from cdp.actions.evm.transfer import account_transfer_strategy, transfer

        return await transfer(
            api_clients=self.__api_clients,
            from_account=self,
            to=to,
            amount=amount,
            token=token,
            network=network,
            transfer_strategy=account_transfer_strategy,
        )

    async def swap(self, options: SwapOptions) -> SwapResult:
        """Swap tokens from one asset to another.

        Args:
            options: The swap options. Must be a SwapOptions instance containing either:
                - Direct swap parameters (network, from_token, to_token, from_amount, slippage_bps)
                - A pre-created swap quote (swap_quote)

                Note: The account address is always used as the taker.

        Returns:
            SwapResult: The result of the swap transaction.

        Examples:
            **Direct swap with parameters**:
            >>> from cdp.actions.evm.swap import SwapOptions
            >>> result = await account.swap(
            ...     SwapOptions(
            ...         network="base",
            ...         from_token="0x833589fCD6eDb6E08f4c7C32D4f71b54bdA02913",  # USDC
            ...         to_token="0x4200000000000000000000000000000000000006",  # WETH
            ...         from_amount="100000000",  # 100 USDC
            ...         slippage_bps=100  # 1% slippage
            ...     )
            ... )

            **Swap with pre-created quote**:
            >>> # First create a quote
            >>> quote = await cdp.evm.create_swap_quote(
            ...     from_token="0x833589fCD6eDb6E08f4c7C32D4f71b54bdA02913",
            ...     to_token="0x4200000000000000000000000000000000000006",
            ...     from_amount="100000000",
            ...     network="base",
            ...     taker=account.address
            ... )
            >>> # Then execute with the quote
            >>> result = await account.swap(
            ...     SwapOptions(swap_quote=quote)
            ... )

        """
        # Validate that options is a SwapOptions instance
        if not isinstance(options, SwapOptions):
            raise TypeError("swap() expects a SwapOptions instance")

        from cdp.actions.evm.swap import AccountSwapStrategy, swap
        from cdp.actions.evm.swap.types import SwapUnavailableResult
        from cdp.evm_client import EvmClient

        # Handle pre-created quote
        if options.swap_quote is not None:
            return await swap(
                api_clients=self.__api_clients,
                from_account=self,
                swap_options=options,
                swap_strategy=AccountSwapStrategy(),
            )

        # Handle direct parameters - create quote first
        # Create an EVM client instance
        evm_client = EvmClient(self.__api_clients)

        # Create the swap quote from parameters, always using account address as taker
        swap_quote = await evm_client.create_swap_quote(
            from_token=options.from_token,
            to_token=options.to_token,
            from_amount=options.from_amount,
            network=options.network,
            taker=self.address,  # Always use account address as taker
            slippage_bps=options.slippage_bps,
            signer_address=self.address,  # Pass account address as signer
        )

        # Check if liquidity is unavailable
        if isinstance(swap_quote, SwapUnavailableResult):
            raise ValueError("Swap unavailable: Insufficient liquidity")

        # Execute the swap
        return await swap(
            api_clients=self.__api_clients,
            from_account=self,
            swap_options=SwapOptions(swap_quote=swap_quote),
            swap_strategy=AccountSwapStrategy(),
        )

    async def quote_swap(
        self,
        from_token: str,
        to_token: str,
        from_amount: str | int,
        network: str,
        slippage_bps: int | None = None,
        signer_address: str | None = None,
    ) -> "QuoteSwapResult":
        """Get a quote for swapping tokens.

        This is a convenience method that calls the underlying create_swap_quote
        with the account's address as the taker.

        Args:
            from_token: The contract address of the token to swap from
            to_token: The contract address of the token to swap to
            from_amount: The amount to swap from (in smallest unit)
            network: The network to execute the swap on
            slippage_bps: Maximum slippage in basis points (100 = 1%). Defaults to 100.
            signer_address: The address that will sign the transaction (for smart accounts). Currently unused.

        Returns:
            QuoteSwapResult: The swap quote with transaction data

        Examples:
            >>> # Get a quote for swapping USDC to WETH (account as taker)
            >>> quote = await account.quote_swap(
            ...     from_token="0x833589fCD6eDb6E08f4c7C32D4f71b54bdA02913",  # USDC
            ...     to_token="0x4200000000000000000000000000000000000006",  # WETH
            ...     from_amount="100000000",  # 100 USDC
            ...     network="base"
            ... )
            >>> print(f"Expected output: {quote.to_amount}")
            >>>
            >>> # Execute the quote if satisfied
            >>> result = await account.swap(SwapOptions(swap_quote=quote))

        """
        from cdp.evm_client import EvmClient

        # Create an EVM client instance
        evm_client = EvmClient(self.__api_clients)

        # Call create_swap_quote with the account address as taker
        return await evm_client.create_swap_quote(
            from_token=from_token,
            to_token=to_token,
            from_amount=from_amount,
            network=network,
            taker=self.address,
            slippage_bps=slippage_bps,
            signer_address=signer_address,
        )

    async def request_faucet(
        self,
        network: str,
        token: str,
    ) -> str:
        """Request a token from the faucet.

        Args:
            network (str): The network to request the faucet for.
            token (str): The token to request the faucet for.

        Returns:
            str: The transaction hash of the faucet request.

        """
        return await request_faucet(
            self.__api_clients.faucets,
            self.address,
            network,
            token,
        )

    async def sign_typed_data(
        self,
        domain: EIP712Domain,
        types: dict[str, Any],
        primary_type: str,
        message: dict[str, Any],
        idempotency_key: str | None = None,
    ) -> str:
        """Sign an EVM typed data.

        Args:
            domain (EIP712Domain): The domain of the message.
            types (Dict[str, Any]): The types of the message.
            primary_type (str): The primary type of the message.
            message (Dict[str, Any]): The message to sign.
            idempotency_key (str, optional): The idempotency key. Defaults to None.

        Returns:
            str: The signature.

        """
        eip712_message = EIP712Message(
            domain=domain,
            types=types,
            primary_type=primary_type,
            message=message,
        )
        response = await self.__evm_accounts_api.sign_evm_typed_data(
            address=self.address,
            eip712_message=eip712_message,
            x_idempotency_key=idempotency_key,
        )
        return response.signature

    async def list_token_balances(
        self,
        network: str,
        page_size: int | None = None,
        page_token: str | None = None,
    ) -> ListTokenBalancesResult:
        """List the token balances for the account on the given network.

        Args:
            network (str): The network to list the token balances for.
            page_size (int, optional): The number of token balances to return per page. Defaults to None.
            page_token (str, optional): The token for the next page of token balances, if any. Defaults to None.

        Returns:
            [ListTokenBalancesResult]: The token balances for the account on the network.

        """
        return await list_token_balances(
            self.__api_clients.evm_token_balances,
            self.address,
            network,
            page_size,
            page_token,
        )

    async def send_transaction(
        self,
        transaction: str | TransactionRequestEIP1559 | DynamicFeeTransaction,
        network: str,
        idempotency_key: str | None = None,
    ) -> str:
        """Send an EVM transaction.

        Args:
            transaction (str | TransactionDictType | DynamicFeeTransaction): The transaction to send.

                This can be either an RLP-encoded transaction to sign and send, as a 0x-prefixed hex string, or an EIP-1559 transaction request object.

                Use TransactionRequestEIP1559 if you would like Coinbase to manage the nonce and gas parameters.

                You can also use DynamicFeeTransaction from eth-account, but you will have to set the nonce and gas parameters manually.

                These are the fields that can be contained in the transaction object:

                    - `to`: (Required) The address of the contract or account to send the transaction to.
                    - `value`: (Optional) The amount of ETH, in wei, to send with the transaction.
                    - `data`: (Optional) The data to send with the transaction; only used for contract calls.
                    - `gas`: (Optional) The amount of gas to use for the transaction.
                    - `nonce`: (Optional) The nonce to use for the transaction. If not provided, the API will assign a nonce to the transaction based on the current state of the account.
                    - `maxFeePerGas`: (Optional) The maximum fee per gas to use for the transaction. If not provided, the API will estimate a value based on current network conditions.
                    - `maxPriorityFeePerGas`: (Optional) The maximum priority fee per gas to use for the transaction. If not provided, the API will estimate a value based on current network conditions.
                    - `accessList`: (Optional) The access list to use for the transaction.
                    - `chainId`: (Ignored) The value of the `chainId` field in the transaction is ignored.
                    - `from`: (Ignored) Ignored in favor of the account address that is sending the transaction.
                    - `type`: (Ignored) The transaction type must always be 0x2 (EIP-1559).

            network (str): The network.
            idempotency_key (str, optional): The idempotency key. Defaults to None.

        Returns:
            str: The transaction hash.

        """
        return await send_transaction(
            evm_accounts=self.__evm_accounts_api,
            address=self.address,
            transaction=transaction,
            network=network,
            idempotency_key=idempotency_key,
        )

    async def quote_fund(
        self,
        network: Literal["base"],
        amount: int,
        token: Literal["eth", "usdc"],
    ) -> Quote:
        """Quote a fund operation.

        Args:
            network: The network to fund the account on.
            amount: The amount of the token to fund in atomic units (e.g. 1000000 for 1 USDC).
            token: The token to fund.

        Returns:
            Quote: A quote object containing:
                - quote_id: The ID of the quote
                - network: The network the quote is for
                - fiat_amount: The amount in fiat currency
                - fiat_currency: The fiat currency (e.g. "usd")
                - token_amount: The amount of tokens to receive
                - token: The token to receive
                - fees: List of fees associated with the quote

        """
        fund_options = QuoteFundOptions(
            network=network,
            amount=amount,
            token=token,
        )

        return await quote_fund(
            api_clients=self.__api_clients,
            address=self.address,
            quote_fund_options=fund_options,
        )

    async def fund(
        self,
        network: Literal["base"],
        amount: int,
        token: Literal["eth", "usdc"],
    ) -> FundOperationResult:
        """Fund an EVM account.

        Args:
            network: The network to fund the account on.
            amount: The amount of the token to fund in atomic units (e.g. 1000000 for 1 USDC).
            token: The token to fund.

        Returns:
            FundOperationResult: The result of the fund operation containing:
                - transfer: A Transfer object with details about the transfer including:
                    - id: The transfer ID
                    - status: The status of the transfer (e.g. "pending", "completed", "failed")
                    - source_amount: The amount in source currency
                    - source_currency: The source currency
                    - target_amount: The amount in target currency
                    - target_currency: The target currency
                    - fees: List of fees associated with the transfer

        """
        fund_options = FundOptions(
            network=network,
            amount=amount,
            token=token,
        )

        return await fund(
            api_clients=self.__api_clients,
            address=self.address,
            fund_options=fund_options,
        )

    async def wait_for_fund_operation_receipt(
        self,
        transfer_id: str,
        timeout_seconds: float = 900,
        interval_seconds: float = 1,
    ) -> Transfer:
        """Wait for a fund operation to complete.

        Args:
            transfer_id: The ID of the transfer to wait for.
            timeout_seconds: The maximum time to wait for completion in seconds. Defaults to 900 (15 minutes).
            interval_seconds: The time between status checks in seconds. Defaults to 1.

        Returns:
            Transfer: The completed transfer object containing:
                - id: The transfer ID
                - status: The final status of the transfer ("completed" or "failed")
                - source_amount: The amount in source currency
                - source_currency: The source currency
                - target_amount: The amount in target currency
                - target_currency: The target currency
                - fees: List of fees associated with the transfer

        Raises:
            TimeoutError: If the transfer does not complete within the timeout period.

        """
        return await wait_for_fund_operation_receipt(
            api_clients=self.__api_clients,
            transfer_id=transfer_id,
            timeout_seconds=timeout_seconds,
            interval_seconds=interval_seconds,
        )

    def __str__(self) -> str:
        """Return a string representation of the EthereumAccount object.

        Returns:
            str: A string representation of the EthereumAccount.

        """
        return f"Ethereum Account Address: {self.address}"

    def __repr__(self) -> str:
        """Return a string representation of the EthereumAccount object.

        Returns:
            str: A string representation of the EthereumAccount.

        """
        return str(self)

    @classmethod
    def to_evm_account(cls, address: str, name: str | None = None) -> "EvmServerAccount":
        """Construct an existing EvmAccount by its address and the name.

        Args:
            address (str): The address of the EvmAccount to retrieve.
            name (str | None): The name of the EvmAccount.

        Returns:
            EvmAccount: The retrieved EvmAccount object.

        Raises:
            Exception: If there's an error retrieving the EvmAccount.

        """
        return cls(address, name)


class ListEvmAccountsResponse(BaseModel):
    """Response model for listing EVM accounts."""

    accounts: list[EvmServerAccount] = Field(description="List of EVM server accounts.")
    next_page_token: str | None = Field(
        None,
        description="Token for the next page of results. If None, there are no more results.",
    )

    model_config = ConfigDict(arbitrary_types_allowed=True)



import base64

from solana.rpc.api import Client as SolanaClient
from solana.rpc.types import TxOpts
from solders.message import Message
from solders.pubkey import Pubkey as PublicKey
from solders.system_program import TransferParams, transfer as solders_transfer
from spl.token.constants import TOKEN_PROGRAM_ID
from spl.token.instructions import (
    TransferCheckedParams,
    get_associated_token_address,
    transfer_checked,
)

from cdp.actions.solana.types import TransferOptions
from cdp.actions.solana.utils import get_or_create_connection, get_usdc_mint_address
from cdp.api_clients import ApiClients
from cdp.openapi_client.models.sign_solana_transaction_request import SignSolanaTransactionRequest


async def transfer(
    api_clients: ApiClients,
    options: TransferOptions,
) -> str:
    """Transfer an amount of a token from an account to another account.

    Args:
        api_clients: The API clients to use to send the transaction
        options: The options for the transfer

    Returns:
        The result of the transfer

    """
    connection = get_or_create_connection(options.network)

    if options.token == "sol":
        tx_bytes = await get_native_transfer(
            connection, options.from_account, options.to_account, options.amount
        )
    else:
        mint_address = (
            get_usdc_mint_address(options.network)
            if options.token == "usdc"
            else options.mint_address
        )
        tx_bytes = get_spl_transfer(
            connection,
            options.from_account,
            options.to_account,
            mint_address,
            options.amount,
        )

    serialized_tx = base64.b64encode(tx_bytes).decode("utf-8")
    response = await api_clients.solana_accounts.sign_solana_transaction(
        address=options.from_account,
        sign_solana_transaction_request=SignSolanaTransactionRequest(
            transaction=serialized_tx,
        ),
    )
    signed_tx = response.signed_transaction
    decoded_signed_tx = base64.b64decode(signed_tx)

    tx_resp = connection.send_raw_transaction(
        decoded_signed_tx, opts=TxOpts(skip_preflight=False, preflight_commitment="processed")
    )

    return tx_resp.value


async def get_native_transfer(
    connection: SolanaClient,
    from_account: str,
    to_account: str,
    amount: int,
) -> bytes:
    """Get the trasnaction bytes for a native transfer."""
    source_pubkey = PublicKey.from_string(from_account)
    dest_pubkey = PublicKey.from_string(to_account)

    blockhash_resp = connection.get_latest_blockhash()
    blockhash = blockhash_resp.value.blockhash

    transfer_params = TransferParams(
        from_pubkey=source_pubkey, to_pubkey=dest_pubkey, lamports=amount
    )
    transfer_instr = solders_transfer(transfer_params)

    message = Message.new_with_blockhash(
        [transfer_instr],
        source_pubkey,
        blockhash,
    )

    # Create a transaction envelope with signature space
    sig_count = bytes([1])  # 1 byte for signature count (1)
    empty_sig = bytes([0] * 64)  # 64 bytes of zeros for the empty signature
    message_bytes = bytes(message)  # Get the serialized message bytes

    # Concatenate to form the transaction bytes
    tx_bytes = sig_count + empty_sig + message_bytes

    return tx_bytes


def get_spl_transfer(
    connection: SolanaClient,
    from_account: str,
    to_account: str,
    mint_address: str,
    amount: int,
) -> bytes:
    """Get the transaction bytes for an SPL token transfer.

    Args:
        connection: The Solana client connection
        from_account: The source account public key
        to_account: The destination account public key
        mint_address: The token mint address
        amount: The amount to transfer

    Returns:
        The transaction bytes

    """
    from_pubkey = PublicKey.from_string(from_account)
    to_pubkey = PublicKey.from_string(to_account)
    mint_pubkey = PublicKey.from_string(mint_address)

    try:
        mint_info = connection.get_account_info(mint_pubkey)
        if not mint_info:
            raise ValueError(f"Failed to fetch mint info for mint address {mint_address}")
    except Exception as e:
        raise ValueError(
            f"Failed to fetch mint info for mint address {mint_address}. Error: {e!s}"
        ) from e

    # Get the decimals from the mint info
    decimals = mint_info.value.data[
        44
    ]  # The decimals are stored at offset 44 in the mint account data

    source_ata = get_associated_token_address(from_pubkey, mint_pubkey)
    destination_ata = get_associated_token_address(to_pubkey, mint_pubkey)

    instructions = []

    try:
        balance = connection.get_token_account_balance(source_ata)

        if int(balance.value.amount) < amount:
            raise ValueError(
                f"Insufficient token balance. Have {balance.value.amount}, need {amount}"
            )
    except Exception as e:
        raise ValueError(f"Error checking source account: {e!s}") from e

    # Check if destination account exists, if not create it
    try:
        dest_account = connection.get_account_info(destination_ata)
        if not dest_account:
            from spl.token.instructions import create_associated_token_account

            instructions.append(
                create_associated_token_account(
                    payer=from_pubkey, owner=to_pubkey, mint=mint_pubkey
                )
            )
    except Exception as e:
        raise ValueError(f"Error checking destination account: {e!s}") from e

    instructions.append(
        transfer_checked(
            params=TransferCheckedParams(
                program_id=TOKEN_PROGRAM_ID,
                amount=amount,
                decimals=decimals,
                dest=destination_ata,
                owner=from_pubkey,
                source=source_ata,
                mint=mint_pubkey,
            ),
        )
    )

    blockhash_resp = connection.get_latest_blockhash()
    blockhash = blockhash_resp.value.blockhash

    message = Message.new_with_blockhash(
        instructions,
        from_pubkey,
        blockhash,
    )

    # Create transaction bytes
    sig_count = bytes([1])  # 1 byte for signature count (1)
    empty_sig = bytes([0] * 64)  # 64 bytes of zeros for the empty signature
    message_bytes = bytes(message)  # Get the serialized message bytes

    # Concatenate to form the transaction bytes
    tx_bytes = sig_count + empty_sig + message_bytes

    return tx_bytes



from unittest.mock import AsyncMock

import pytest

from cdp.actions.solana.sign_message import sign_message
from cdp.openapi_client.api.solana_accounts_api import SolanaAccountsApi
from cdp.openapi_client.models.sign_solana_message200_response import (
    SignSolanaMessage200Response as SignSolanaMessageResponse,
)
from cdp.openapi_client.models.sign_solana_message_request import SignSolanaMessageRequest


@pytest.mark.asyncio
async def test_sign_message_success():
    """Test successful message signing."""
    # Setup mock Solana accounts API
    mock_solana_accounts_api = AsyncMock(spec=SolanaAccountsApi)

    # Setup test data
    test_address = "14grJpemFaf88c8tiVb77W7TYg2W3ir6pfkKz3YjhhZ5"
    test_message = "Hello, Solana!"
    test_idempotency_key = "test-idempotency-key"
    test_signature = "test_signature"

    # Setup mock response
    mock_response = SignSolanaMessageResponse(signature=test_signature)
    mock_solana_accounts_api.sign_solana_message = AsyncMock(return_value=mock_response)

    # Call the function
    result = await sign_message(
        solana_accounts_api=mock_solana_accounts_api,
        address=test_address,
        message=test_message,
        idempotency_key=test_idempotency_key,
    )

    # Verify the API was called correctly
    mock_solana_accounts_api.sign_solana_message.assert_called_once_with(
        sign_solana_message_request=SignSolanaMessageRequest(
            message=test_message,
        ),
        address=test_address,
        x_idempotency_key=test_idempotency_key,
    )

    # Verify the response
    assert result == mock_response
    assert result.signature == test_signature


@pytest.mark.asyncio
async def test_sign_message_error():
    """Test message signing error handling."""
    # Setup mock Solana accounts API
    mock_solana_accounts_api = AsyncMock(spec=SolanaAccountsApi)

    # Setup test data
    test_address = "14grJpemFaf88c8tiVb77W7TYg2W3ir6pfkKz3YjhhZ5"
    test_message = "Hello, Solana!"
    test_idempotency_key = "test-idempotency-key"

    # Setup mock to raise an error
    mock_solana_accounts_api.sign_solana_message = AsyncMock(side_effect=Exception("API Error"))

    # Verify the error is propagated
    with pytest.raises(Exception) as exc_info:
        await sign_message(
            solana_accounts_api=mock_solana_accounts_api,
            address=test_address,
            message=test_message,
            idempotency_key=test_idempotency_key,
        )

    assert str(exc_info.value) == "API Error"

    # Verify the API was called correctly
    mock_solana_accounts_api.sign_solana_message.assert_called_once_with(
        sign_solana_message_request=SignSolanaMessageRequest(
            message=test_message,
        ),
        address=test_address,
        x_idempotency_key=test_idempotency_key,
    )



from cdp.openapi_client.api.solana_accounts_api import SolanaAccountsApi
from cdp.openapi_client.models.sign_solana_transaction200_response import (
    SignSolanaTransaction200Response as SignSolanaTransactionResponse,
)
from cdp.openapi_client.models.sign_solana_transaction_request import (
    SignSolanaTransactionRequest,
)


async def sign_transaction(
    solana_accounts_api: SolanaAccountsApi,
    address: str,
    transaction: str,
    idempotency_key: str,
) -> SignSolanaTransactionResponse:
    """Sign a transaction.

    Args:
        solana_accounts_api (SolanaAccountsApi): The Solana accounts API.
        address (str): The address of the Solana account.
        transaction (str): The transaction to sign.
        idempotency_key (str): The idempotency key.

    Returns:
        SignSolanaTransactionResponse: The signature of the transaction.

    """
    return await solana_accounts_api.sign_solana_transaction(
        sign_solana_transaction_request=SignSolanaTransactionRequest(
            transaction=transaction,
        ),
        address=address,
        x_idempotency_key=idempotency_key,
    )
